import time as time_lib
import argparse
import os, sys
import joblib
import json
import tensorflow as tf
import pandas
import numpy
import joblib
import tempfile

from easierSDK.easier import EasierSDK
from easierSDK.classes.categories import Categories
from easierSDK.classes.model_metadata import ModelMetadata
from easierSDK.classes.easier_model import EasierModel
import easierSDK.classes.constants as constants
from easierSDK import datasetsAPI

# Variable definition
easier = None
easier_model = None

# Environment variables loading
checkpoint_dir = os.getenv('CHECKPOINT_DIR', '/train/checkpoint')
data_dir = os.getenv('DATA_DIR', '/train/data')
model_dir = os.getenv('MODEL_DIR', '/train/model')
model_name = os.getenv('MODEL_NAME', 'model')

save_freq = os.getenv('SAVE_FREQ', 10)



def build_compile_model(model_dir:str, model_name:str):
    pass
    model = tf.keras.models.model_from_json(json.loads(model_dir + '/' + model_name + '.json'), 
                custom_objects=None)
    model.compile(
        optimizer='rmsprop', 
        loss=tf.keras.losses.categorical_crossentropy,
        metrics=[tf.keras.metrics.mean_squared_error]
    )
    return model

def compile_model(easier_model:EasierModel):
    # TF Compile should be done within distributed strategy
    easier_model.model.compile(
        optimizer=easier_model.training_metadata.optimizer, 
        loss=easier_model.training_metadata.loss,
        metrics=easier_model.training_metadata.metrics
    )
    easier_model.model.summary()
    return easier_model

if __name__ == "__main__":
    
    # Initializations
    easier_user = os.getenv("easier_user")
    easier_password = os.getenv("easier_password")
    
    ignore_cert_warning = os.getenv("ignore_cert_warning")
    ignore_cert_warning = True if str(ignore_cert_warning).lower() == 'true' else False

    minio_region = os.getenv("minio_region")
    
    minio_secure = os.getenv("minio_secure")
    minio_secure = True if str(minio_secure).lower() == 'true' else False
    
    easier_url = os.getenv("easier_url")
    easier = EasierSDK(easier_user=easier_user, easier_password=easier_password, easier_url=easier_url, minio_secure=minio_secure, minio_region=minio_region, ignore_cert_warning=ignore_cert_warning)

    repo = os.getenv("repo")
    category = os.getenv("category")
    model_name = os.getenv("model_name")
    experimentID = os.getenv("experimentID")
    data_path = os.getenv("data_path")

    easier_model = easier.models.get_model(repo_name=repo, category=category, model_name=model_name, experimentID=experimentID, load_level=constants.FULL, training=True)

    x, y, x_test, y_test = easier.training.get_training_data(repo_path=data_path)

    easier_model = compile_model(easier_model)

    history = easier_model.model.fit(x, y, validation_data=(x_test, y_test), epochs=easier_model.training_metadata.epochs, batch_size=easier_model.training_metadata.batch_size, verbose=2)

    easier.training.clear_training_data(repo_path=data_path, repo_name=repo)

    easier.models._finish_model_training(easier_model, repo_name=repo, category=category, experimentID=experimentID)

    # Successfull exit
    sys.exit()



#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import sys
sys.path.append('..')
sys.path.append('.')

import time
import json

from easierSDK.classes.categories import Categories

class CategoryMetadata():
    """Class to store the repository's metadata information.
    """

    name = ''
    models = {}
    num_models = 0
    datasets = {}
    num_datasets = 0

    def __init__(self, name:str, models:dict, num_models:int, datasets:dict, num_datasets:int):
        """Constructor for RepositoryMetadata class.

        Args:
            name (str): Repository name
            models (list): List of ModelMetadata objects of models under this repository
            num_models (int): Number of models in list
            datasets (list): List of DataSetMetadata objects of datasets under this repository
            num_datasets (int): Number of datasets in list
        """
        self.name = name
        self.models = models
        self.datasets = datasets
        self.num_models = num_models
        self.num_datasets = num_datasets


    def print_models(self):
        """Print the metadata information.
        """
        row_format ="{:<30}" * 4

        print("MODELS:")
        print(row_format.format(*['Name', 'Category', 'Last Modification', 'Num Experiments']))
        for model_metadata in self.models:
            print(row_format.format(*[self.models[model_metadata].name, 
                self.models[model_metadata].category, self.models[model_metadata].last_modified, 
                str(len(self.models[model_metadata].experimentIDs))]))

    def print_datasets(self):
        """Print the metadata information.
        """
        row_format ="{:<30}" * 3
        
        print("DATASETS:")
        print(row_format.format(*['Name', 'Category', 'Last Modification']))
        for dataset_metadata in self.datasets:
            print(row_format.format(*[self.datasets[dataset_metadata].name, self.datasets[dataset_metadata].category, self.datasets[dataset_metadata].last_modified]))
   
    def pretty_print(self):
        """Print the metadata information.
        """
        row_format ="{:<30}" * 4

        print("MODELS:")
        print(row_format.format(*['Name', 'Category', 'Last Modification', 'Num Experiments']))
        for model_metadata in self.models:
            print(row_format.format(*[self.models[model_metadata].name, 
                self.models[model_metadata].category, self.models[model_metadata].last_modified, 
                str(len(self.models[model_metadata].experimentIDs))]))
        
        row_format ="{:<30}" * 3
        print("\nDATASETS:")
        print(row_format.format(*['Name', 'Category', 'Last Modification']))
        for dataset_metadata in self.datasets:
            print(row_format.format(*[self.datasets[dataset_metadata].name, self.datasets[dataset_metadata].category, self.datasets[dataset_metadata].last_modified]))




ingress = """
apiVersion: networking.k8s.io/v1beta1
kind: Ingress
metadata:
  name: ingress-easier-training
  kubernetes.io/ingress.allow-http: "true"
  nginx.ingress.kubernetes.io/backend-protocol: HTTP
  nginx.ingress.kubernetes.io/secure-backends: "false"
  labels:
    app: easier-training
spec:
  tls:
  - secretName: wildcard-easier-ai.eu
  rules:
  - host: easier-ai.eu
    http:
      paths:
      - backend:
          serviceName: service-easier-training
          servicePort: 5100
"""

ingress_dist = """
apiVersion: networking.k8s.io/v1beta1
kind: Ingress
metadata:
  name: ingress-easier-dist-training
  kubernetes.io/ingress.allow-http: "true"
  nginx.ingress.kubernetes.io/backend-protocol: HTTP
  nginx.ingress.kubernetes.io/secure-backends: "false"
  labels:
    app: easier-dist-training
spec:
  tls:
  - secretName: wildcard-easier-ai.eu
  rules:
  - host: easier-ai.eu
    http:
      paths:
      - backend:
          serviceName: service-easier-dist-training
          servicePort: 6006
"""


ingress_dist_v1 = """
apiVersion: networking.k8s.io/v1
kind: Ingress
metadata:
  name: ingress-easier-dist-training
  # kubernetes.io/ingress.allow-http: "true"
  # nginx.ingress.kubernetes.io/backend-protocol: HTTP
  # nginx.ingress.kubernetes.io/secure-backends: "false"
  labels:
    app: easier-dist-training
spec:
  tls:
  - secretName: wildcard-easier-ai.eu
  rules:
  - host: easier-ai.eu
    http:
      paths:
      - path: /
        pathType: Prefix
        backend:
          service:
            name: service-easier-dist-training
            port: 
              number: 6006
"""


import os
import kubernetes
import yaml
from kubernetes import client, config
from kubernetes.client.rest import ApiException
import string
import random

from easierSDK.classes.categories import Categories
from easierSDK.serving import easier_serving_config_map
from easierSDK.serving import easier_serving_deployment
from easierSDK.serving import easier_serving_service 
from easierSDK.serving import easier_serving_ingress 

class ServingAPI():
    """Class to control the Serving API of EasierSDK.
    """

    def __init__(self, easier_user, easier_password, minio_client, my_public_repo, my_private_repo, easier_instance=None):
        """Constructor for the ServingAPI.

        Args:
            minio_client (Minio): Minio client object with user session initialized.
            my_public_repo (str): Name of the public bucket of the user.
            my_private_repo (str): Name of the private bucket of the user.
        """
        self._easier_user = easier_user
        self._easier_password = easier_password
        self.minio_client = minio_client
        self.my_public_repo = my_public_repo
        self.my_private_repo = my_private_repo
        self.namespace = None
        self._easier_instance = easier_instance
        
    def id_generator(self, size=16, chars=string.ascii_lowercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))

    def create_model_serving(self, repo_name:str, category:Categories, model_name:str, experimentID:int, base_hostname_url='.easier-ai.eu', resource_name=None, serving_name=None):
        """Creates a model serving deployment on the infrastructure.

        Args:
            repo_name (str): name of the repository where the model is
            category (Categories): category of the model
            model_name (str): name of the model
            experimentID (int): version of the model
            namespace ([type], optional): namespace where the deployment will be done. Defaults to None to take user's namespace.

        Returns:
            str: hostname of the deployed model
        """
        
        if isinstance(category, str):
           category = Categories[category.upper()]
        
        # Test if model can be loaded. 
        easier_model = self._easier_instance.models.get_model(repo_name=repo_name, category=category, model_name=model_name, experimentID=experimentID)    
        if easier_model.get_model() is None:
            print("ERROR: Could not load model " + str(model_name) +
                    " from repository " + repo_name + " and category " + category.value)
            return None

        if self.namespace is None:
            print("ERROR: No context (namespace) was provided. Please, initialize the platform connection before creating model servings.")
            return None
        namespace = self.namespace  
        
        # Enter a context with an instance of the API kubernetes.client
        with kubernetes.client.ApiClient() as api_client:
            # Create an instance of the API class
            api_instance = kubernetes.client.CoreV1Api(api_client)
            
            if resource_name is None: 
                random_name = self.id_generator(size=5)
            else:
                random_name = resource_name
            
            # config_map = None
            # with open(os.path.join(os.path.dirname(__file__), "easier-serving-config_map.yaml")) as f:
            config_map = yaml.safe_load(easier_serving_config_map.config_map)

            config_map['metadata']['name'] += '-' + random_name
            config_map['data']['easier_user'] = self._easier_user
            config_map['data']['easier_password'] = self._easier_password
            config_map['data']['repo'] = repo_name
            config_map['data']['category'] = category.value
            config_map['data']['model_name'] = model_name
            config_map['data']['experimentID'] = str(experimentID)
            config_map['data']['ignore_cert_warning'] = str(self._easier_instance.ignore_cert_warning)
            config_map['data']['minio_region'] = self._easier_instance.minio_region
            config_map['data']['minio_secure'] = str(self._easier_instance.minio_secure)
            config_map['data']['easier_url'] = self._easier_instance.easier_url
                                    
            try:
                api_response = api_instance.create_namespaced_config_map(namespace, config_map, pretty='true')
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->create_namespaced_config_map: %s\n" % e)

            # deployment = None
            # with open(os.path.join(os.path.dirname(__file__), "easier-serving-deployment.yaml")) as f:
            deployment = yaml.safe_load(easier_serving_deployment.deployment)
        
            deployment['metadata']['name'] += '-' + random_name
            deployment['metadata']['labels']['app'] += '-' +  random_name
            deployment['spec']['selector']['matchLabels']['app'] += '-' + random_name
            deployment['spec']['template']['metadata']['labels']['app'] += '-' + random_name
            deployment['spec']['containers'][0]['name'] += '-' + random_name
            deployment['spec']['containers'][0]['envFrom'][0]['configMapRef']['name'] += '-' + random_name
    
            try:
                api_response = api_instance.create_namespaced_pod(namespace, deployment, pretty='true')
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->create_namespaced_pod: %s\n" % e)

            # service = None
            # with open(os.path.join(os.path.dirname(__file__), "easier-serving-service.yaml")) as f:
            service = yaml.safe_load(easier_serving_service.service)
        
            service['metadata']['name'] += '-' + random_name
            service['metadata']['labels']['app'] += '-' + random_name
            service['spec']['selector']['app'] += '-' + random_name

            try:
                api_response = api_instance.create_namespaced_service(namespace, service, pretty='true')
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->create_namespaced_service: %s\n" % e)
        
            networking_v1_beta1_api = kubernetes.client.NetworkingV1beta1Api()
            hostname = None
            # ingress = None
            # with open(os.path.join(os.path.dirname(__file__), "easier-serving-ingress.yaml")) as f:
            ingress = yaml.safe_load(easier_serving_ingress.ingress)
        
            ingress['metadata']['name'] += '-' + random_name
            ingress['metadata']['labels']['app'] += '-' + random_name
            if serving_name:
                ingress['spec']['rules'][0]['host'] = serving_name + base_hostname_url
            else:
                ingress['spec']['rules'][0]['host'] = model_name + '-' + 'easier-serving' + '-' + random_name + base_hostname_url
            ingress['spec']['rules'][0]['http']['paths'][0]['backend']['serviceName'] += '-' + random_name

            try:
                api_response = networking_v1_beta1_api.create_namespaced_ingress(namespace, ingress, pretty='true')
                # print(api_response)
                hostname = ingress['spec']['rules'][0]['host']
            except ApiException as e:
                print("Exception when calling CoreV1Api->create_namespaced_ingress: %s\n" % e)
                        
            if hostname:
                print("Your model will be served shortly in: " + str(hostname))
                print("The identifier for the deployment is: " + random_name)
            else:
                print("There was a problem serving your model")

        return hostname

    def delete_serving(self, deployment_identifier:str):
        """Deletes the deployment resources of a serving.

        Args:
            hostname (str): hostname given as identifier when a serving was created
        """
        
        if self.namespace is None:
            print("ERROR: No context (namespace) was provided. Please, initialize the platform connection before creating model servings.")
            return None
        namespace = self.namespace  
        
        # Enter a context with an instance of the API kubernetes.client
        with kubernetes.client.ApiClient() as api_client:
            # Create an instance of the API class
            api_instance = kubernetes.client.CoreV1Api(api_client)

            try:
                config_map = yaml.safe_load(easier_serving_config_map.config_map)
            
                config_map['metadata']['name'] += '-' + deployment_identifier
                api_response = api_instance.delete_namespaced_config_map(config_map, namespace)
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->delete_namespaced_config_map: %s\n" % e)

            try:
                deployment = yaml.safe_load(easier_serving_deployment.deployment)
            
                deployment['metadata']['name'] += '-' + deployment_identifier
                api_response = api_instance.delete_namespaced_pod(deployment, namespace)
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->delete_namespaced_pod: %s\n" % e)
            
            try:
                service = yaml.safe_load(easier_serving_service.service)
            
                service['metadata']['name'] += '-' + deployment_identifier
                api_response = api_instance.delete_namespaced_service(service, namespace)
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->delete_namespaced_pod: %s\n" % e)

            try:
                ingress = yaml.safe_load(easier_serving_ingress.ingress)
            
                ingress['metadata']['name'] += '-' + deployment_identifier
                
                networking_v1_beta1_api = kubernetes.client.NetworkingV1beta1Api()
                api_response = networking_v1_beta1_api.delete_namespaced_ingress(ingress, namespace)
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->delete_namespaced_pod: %s\n" % e)
    


config_map = """
apiVersion: v1
kind: ConfigMap
metadata:
  name: easier-training
data:
  easier_user: 
  easier_password: 
  repo: 
  category: 
  model_name: 
  experimentID: 
  data_path:
  TF_CONFIG:
"""

config_map_distributed = """
apiVersion: v1
kind: ConfigMap
metadata:
  name: easier-dist-training
data:
  easier_user: 
  easier_password: 
  repo: 
  category: 
  model_name: 
  experimentID: 
  data_path:
  TF_CONFIG:
"""


#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import sys
sys.path.append('..')
sys.path.append('.')

import time
import json
import numpy

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

class TrainingMetadata():
    """Class to store the training's metadata information.
    """

    inference_mode = ''
    optimizer = ''
    loss = ''
    epochs = 100
    batch_size = 10
    metrics = []
    

    def __init__(self, f:dict=None):
        """Constructor for the Training Metadata Class.

        Args:
            f (dict, optional): Dictionary with the Training's metadata information.
        """
        if f is not None:
            if 'inference_mode' in f: self.inference_mode = f['inference_mode']
            if 'optimizer' in f: self.optimizer = f['optimizer']
            if 'loss' in f: self.loss = f['loss']
            if 'epochs' in f: self.epochs = f['epochs']
            if 'batch_size' in f: self.batch_size = f['batch_size']
            if 'metrics' in f: self.metrics = f['metrics']
        else:
            pass

    # def seria
    def pretty_print(self):
        """Print the training metadata information.
        """
        row_format ="{:<30}" * 2
        print(row_format.format(*['Inference Mode:', self.inference_mode]))
        print(row_format.format(*['Optimizer:', self.optimizer]))
        print(row_format.format(*['Loss function:', self.loss]))
        print(row_format.format(*['Epochs:', self.epochs]))
        print(row_format.format(*['Batch size:', str(self.batch_size)]))
        print(row_format.format(*['Metrics:', str(self.metrics)]))
        
    def dump_to_file(self, path:str):
        """Dump metadata information to a file.

        Args:
            path (str): Path to store the file with the metadata information.
        """
        with open(path+'/training_metadata.json', 'w') as f:
            f.write(json.dumps(self.__dict__, default=lambda o: str(o.value), 
                sort_keys=True, indent=4, cls=NumpyEncoder))




#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

from enum import Enum

class Categories(Enum):
    """Categories under which EASIER Models and Datasets are classified.
    """

    HEALTH = 'health'
    TRANSPORT = 'transport'
    SECURITY = 'security'
    AIRSPACE = 'airspace'
    EDUCATION = 'education'
    MISC = 'misc'


#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import minio
from minio import Minio
import os
import json
import certifi
import urllib3
import kubernetes
from kubernetes.client.rest import ApiException
import yaml
from tqdm import tqdm

from easierSDK import modelsAPI, datasetsAPI
from easierSDK.classes.categories import Categories
from easierSDK.classes.repository_metadata import RepositoryMetadata
from easierSDK.classes.model_metadata import ModelMetadata
from easierSDK.classes.dataset_metadata import DatasetMetadata
from easierSDK.classes.category_metadata import CategoryMetadata

from easierSDK.serving import servingAPI
from easierSDK.training import trainingAPI, distributed_trainingAPI

import urllib3
urllib3.disable_warnings()

class EasierSDK():
    """Higher level class to interact with the EASIER platform.
    """

    _MODELS = 'models'
    _DATASETS = 'datasets'
    
    def __init__(self, easier_user:str, easier_password:str, easier_url:str="minio.easier-ai.eu", minio_secure=True, minio_region='es', ignore_cert_warning=False):
        """Initializer for the class.

        Args:
            easier_url (str): URL to an EASIER MINIO deployment.
            easier_user (str): Username.
            easier_password (str): Password.
        """
        self.my_easier_user = easier_user
        self.ignore_cert_warning = ignore_cert_warning
        self.minio_region = minio_region
        self.minio_secure = minio_secure
        self.easier_url = easier_url

        if ignore_cert_warning:
            cert = 'CERT_NONE'
        else: 
            cert = 'CERT_REQUIRED'

        # Load CA certificates from SSL_CERT_FILE file if set
        ca_certs = os.environ.get('SSL_CERT_FILE') or certifi.where()
        _http = urllib3.PoolManager(
            timeout=1,
            maxsize=minio.helpers.MAX_POOL_SIZE,
            cert_reqs=cert,
            ca_certs=ca_certs,
            retries=urllib3.Retry(
                total=3,
                backoff_factor=0.2,
                status_forcelist=[500, 502, 503, 504]
            )
        )

        self.minio_client = Minio(easier_url, access_key=easier_user, secret_key=easier_password, secure=minio_secure, region=minio_region, http_client=_http)
        
        # Test Connection
        try:
            self.minio_client.list_buckets()
        except Exception as e:
            raise Exception("ERROR when connecting to EASIER Repositories. Check input parameters.")

        # IF test is ok, then we reinit with default parameters of the http lib
        _http = urllib3.PoolManager(
            timeout=urllib3.Timeout.DEFAULT_TIMEOUT,
            maxsize=minio.helpers.MAX_POOL_SIZE,
            cert_reqs=cert,
            ca_certs=ca_certs,
            retries=urllib3.Retry(
                total=5,
                backoff_factor=0.2,
                status_forcelist=[500, 502, 503, 504]
            )
        )

        self.minio_client = Minio(easier_url, access_key=easier_user, secret_key=easier_password, secure=minio_secure, region=minio_region, http_client=_http)
        
        self.my_public_repo = str(easier_user).split('@')[0] + '-public'
        self.my_private_repo = str(easier_user).split('@')[0] + '-private' 

        self.models = modelsAPI.ModelsAPI(self.minio_client, self.my_public_repo, self.my_private_repo)
        self.datasets = datasetsAPI.DatasetsAPI(self.minio_client, self.my_public_repo, self.my_private_repo)

        self.serving = servingAPI.ServingAPI(easier_user, easier_password,  self.minio_client, 
                self.my_public_repo, self.my_private_repo, self)

        self.training = trainingAPI.TrainingAPI(easier_user, easier_password, self.minio_client, 
                self.my_public_repo, self.my_private_repo, self)

        self.distributed_training = distributed_trainingAPI.DistributedTrainingAPI(easier_user, easier_password, self.minio_client, 
                self.my_public_repo, self.my_private_repo, self)

    def initialize_connection(self, kube_config_path:str=None):
        """Initialices the connection to kubernetes infrastructure.

        Args:
            kube_config_path (str, optional): path to user's kubeconfig file. Defaults to None to get the file from ~/.kube/config
        """
        if kube_config_path:
            kubernetes.config.load_kube_config(kube_config_path)
            with open(os.path.join(os.path.dirname(__file__), kube_config_path)) as f:
                kubeconfig = yaml.safe_load(f)
                self.namespace = kubeconfig['contexts'][0]['context']['namespace']
        else:
            kubernetes.config.load_kube_config()
            with open(os.path.join(os.path.dirname(__file__), os.environ.get('KUBECONFIG', '~/.kube/config'))) as f:
                kubeconfig = yaml.safe_load(f)
                self.namespace = kubeconfig['contexts'][0]['context']['namespace']

        print("Current context on namespace: " + str(self.namespace))
        self.serving.namespace = self.namespace
        self.training.namespace = self.namespace
        self.distributed_training.namespace = self.namespace
    

    def _get_resource_names(self, resource_list):
        resource_names = []
        for resource in resource_list:
            resource_names.append(resource.metadata.name)
        return resource_names

    def get_deployed_resources(self):
        """Returns a dictionary with the deployed resources of the user. Fields of the dictionary are:
            pods: a python list of deployed pods names
            services: a python list of deployed services names
            ingress: a python list of deployed ingress names
            jobs: a python list of deployed jobs names

        Returns:
            dict: python dictionary with the deployed resources of the user
        """
        resources_dict = {}
        
        # Enter a context with an instance of the API kubernetes.client
        with kubernetes.client.ApiClient() as api_client:
            # Create an instance of the API class
            api_instance = kubernetes.client.CoreV1Api(api_client)

            try:
                api_response = api_instance.list_namespaced_pod(self.namespace, pretty='true')
                pods_list = self._get_resource_names(api_response.items)
                resources_dict['pods'] = pods_list
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->list_namespaced_pod: %s\n" % e)

            try:
                api_response = api_instance.list_namespaced_service(self.namespace, pretty='true')
                services_list = self._get_resource_names(api_response.items)
                resources_dict['services'] = services_list
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->list_namespaced_service: %s\n" % e)
            
            networking_v1_beta1_api = kubernetes.client.NetworkingV1beta1Api()

            try:
                api_response = networking_v1_beta1_api.list_namespaced_ingress(self.namespace, pretty='true')
                # print(api_response)
                ingress_list = self._get_resource_names(api_response.items)
                resources_dict['ingress'] = ingress_list
            except ApiException as e:
                print("Exception when calling NetworkingV1beta1Api->list_namespaced_ingress: %s\n" % e)

            api_instance = kubernetes.client.BatchV1Api(api_client)

            try:
                api_response = api_instance.list_namespaced_job(self.namespace, pretty='true')
                # print(api_response)
                jobs_list = self._get_resource_names(api_response.items)
                resources_dict['jobs'] = jobs_list
            except ApiException as e:
                print("Exception when calling BatchV1Api->list_namespaced_job: %s\n" % e)
            
            return resources_dict

    def get_repositories_metadata(self, category:Categories=None):
        """Reads repositories information regarding Datasets and Models and returns a dictionary with the information.

        Args:
            category (Categories, optional): Limit search to only one Category. Defaults to None.

        Returns:
            dict: dictionary with the repositories information
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        if category is None:
            categories = [cat for cat in Categories]
        else:
            categories = [category]
        
        repositories = {}
        repo_list = tqdm(self.minio_client.list_buckets(), colour="red", desc="Getting repositories information...")
        
        for repo in repo_list:
            cat_dict = {}
            repo_list.set_postfix({'repository': repo.name})
            for cat in categories: 
                num_models, models_list = self._count_repo_models(repo.name, cat)
                num_datasets, datasets_list = self._count_repo_datasets(repo.name, cat)
            
                cat_models = {}
                for model_path in models_list:
                    minio_path = model_path + '/' + 'metadata.json'
                    filename = minio_path.split('/')[1:]
                    filename = '/'.join(filename)
                    local_file = '/tmp/metadata.json'
                    try:
                        self.minio_client.fget_object(repo.name, filename, local_file)
                    except minio.error.NoSuchKey as ex:
                        continue
                    
                    with open(local_file, 'r') as f:
                        model_metadata = ModelMetadata(json.load(f))
                    if os.path.exists(local_file):
                        os.remove(local_file)
                    
                    _, model_metadata.experimentIDs = self.models._count_model_experiments(repo.name, cat.value, model_metadata.name)
                    cat_models[model_metadata.name] = model_metadata

                cat_datasets = {}
                for dataset_path in datasets_list:
                    minio_path = dataset_path + '/' + 'metadata.json'
                    filename = minio_path.split('/')[1:]
                    filename = '/'.join(filename)
                    local_file = '/tmp/metadata.json'
                    try:
                        self.minio_client.fget_object(repo.name, filename, local_file)
                    except minio.error.NoSuchKey as ex:
                        continue
                    
                    with open(local_file, 'r') as f:
                        dataset_metadata = DatasetMetadata(json.load(f))
                    if os.path.exists(local_file):
                        os.remove(local_file)

                    cat_datasets[dataset_metadata.name] = dataset_metadata

                cat_metadata = CategoryMetadata(cat.value, cat_models, num_models, cat_datasets, num_datasets)
                cat_dict[cat.value] = cat_metadata

            repo_metadata = RepositoryMetadata(repo.name, cat_dict)
            repositories[repo.name] = repo_metadata
        return repositories

    def _count_repo_models(self, repo_name:str, category_name:str=None):
        """Count number of models under a repository.

        Args:
            repo_name (str): Repository to count models from.
            category_name (str): Category to count models from.

        Returns:
            int: Number of models under the repository.
            list: List of models in the repository.
        """
        
        num_models = 0
        models_list = []
        
        if category_name:
            if isinstance(category_name, str):
                category_name = Categories[category_name.upper()]
            categories = [category_name]
        else:
            categories = [cat for cat in Categories]

        for category in categories:
            for obs in self.minio_client.list_objects(repo_name, prefix=self._MODELS + '/' + category.value, recursive=True):
                experiment_model = obs.object_name.split('/')[0:3]
                experiment_model = repo_name + '/' + '/'.join(experiment_model)
                if experiment_model not in models_list:
                    num_models += 1
                    models_list.append(experiment_model)
        return num_models, models_list

    def _count_repo_datasets(self, repo_name:str, category_name:str=None):
        """Count number of datasets under a repository.

        Args:
            repo_name (str): Repository to count datasets from.
            category_name (str): Category to count models from.

        Returns:
            int: Number of datasets under the repository.
            list: List of datasets in the repository.
        """
        num_datasets = 0
        datasets_list = []

        if category_name:
            if isinstance(category_name, str):
                category_name = Categories[category_name.upper()]
            categories = [category_name]
        else:
            categories = [cat for cat in Categories]

        for category in categories:
            for obs in self.minio_client.list_objects(repo_name, prefix=self._DATASETS + '/' + category.value, recursive=True):
                experiment_dataset = obs.object_name.split('/')[0:3]
                experiment_dataset = repo_name + '/' + '/'.join(experiment_dataset)
                if experiment_dataset not in datasets_list:
                    num_datasets += 1
                    datasets_list.append(experiment_dataset)
        return num_datasets, datasets_list
        
    def _count_category_models(self, category:Categories):
        """Count all available models under a specific category.

        Args:
            category (Categories): Category to which count the models from.

        Returns:
            int: number of models under the category
            list: list of models under the category
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        num_models = 0
        models_list = []
        repo_list = self.minio_client.list_buckets()
        for repo in repo_list:
            for obs in self.minio_client.list_objects(repo.name, prefix=self._MODELS + '/' + category.value, recursive=True):
                experiment_model = obs.object_name.split('/')[0:3]
                experiment_model = repo.name + '/' + '/'.join(experiment_model)            
                if experiment_model not in models_list:
                    num_models += 1
                    models_list.append(experiment_model)
        return num_models, models_list
    
    def _count_category_datasets(self, category:Categories):
        """Count all available datasets under a specific category.

        Args:
            category (Categories): Category to which count the datasets from.

        Returns:
            int: number of datasets under the category
            list: list of datasets under the category
        """
        if isinstance(category, str):
           category = Categories[category.upper()]
           
        num_datasets = 0
        datasets_list = []
        repo_list = self.minio_client.list_buckets()
        for repo in repo_list:
            for obs in self.minio_client.list_objects(repo.name, prefix=self._DATASETS + '/' + category.value, recursive=True):
                experiment_dataset = obs.object_name.split('/')[0:3]
                experiment_dataset = repo.name + '/' + '/'.join(experiment_dataset)
                if experiment_dataset not in datasets_list:
                    num_datasets += 1
                    datasets_list.append(experiment_dataset)
        return num_datasets, datasets_list



import time as time_lib
import argparse
from flask_api import status
import flask
from flask_cors import CORS
import os
import joblib
import json
import tensorflow as tf
import pandas
import numpy

app = flask.Flask(__name__)
cors = CORS(app)
app.config['CORS_HEADERS'] = 'Content-Type'

from easierSDK.easier import EasierSDK
from easierSDK.classes.categories import Categories
from easierSDK.classes.model_metadata import ModelMetadata
from easierSDK.classes.easier_model import EasierModel
import easierSDK.classes.constants as constants
from easierSDK import datasetsAPI

# Variable definition
easier = None
easier_model = None

@app.route("/predict" + "/<string:input_data>", methods=["GET"])
def predict(input_data):
    request = flask.request
    results = {"status": "error"}

    if request.method == "GET":
        results["status"] = "success"

        stat = status.HTTP_200_OK

        global easier_model

        try:
            start_time = time_lib.clock()
        except Exception as e:
            start_time = time_lib.process_time()
        
        data = input_data.split(",")
        input_df = pandas.DataFrame([data]).apply(pandas.to_numeric, errors='ignore', axis=0)

        predictions = easier_model.get_model().predict(input_df.values)

        try:
            end_time = time_lib.clock()
        except Exception as e:
            end_time = time_lib.process_time()

        elapsed_time = end_time - start_time

        results["latency"] = elapsed_time
        results["prediction"] = predictions.tolist()
    else:
        results["status"] = "error_method_incorrect"
        stat = status.HTTP_405_METHOD_NOT_ALLOWED

    return flask.jsonify(results), stat 

@app.route("/", methods=["GET"])
@app.route("/compute", methods=["GET"])
def compute():
    request = flask.request
    results = {"status": "error"}
    print(request)
    if request.method == "GET":
        results["status"] = "success"

        stat = status.HTTP_200_OK

        try:
            start_time = time_lib.clock()
        except Exception as e:
            start_time = time_lib.process_time()

        data = request.get_json()
        confidences, classes_pred, predictions = get_prediction(data)

        try:
            end_time = time_lib.clock()
        except Exception as e:
            end_time = time_lib.process_time()

        elapsed_time = end_time - start_time
        results["latency"] = elapsed_time

        if predictions is None:
            stat = status.HTTP_500_INTERNAL_SERVER_ERROR
            results["status"] = "error while predicting"
            return flask.jsonify(results), stat

        results["prediction"] = predictions.tolist()
        if confidences: 
            results["confidence"] = confidences
            results["class"] = classes_pred.tolist()

    else:
        results["status"] = "error_method_incorrect"
        stat = status.HTTP_405_METHOD_NOT_ALLOWED

    return flask.jsonify(results), stat

def get_prediction(features_values_dict):
    
    global easier_model
    global easier

    # Parse data
    try:
        if easier_model.metadata.features and len(easier_model.metadata.features)>0:
            data = pandas.DataFrame(columns=easier_model.metadata.features)
            input_df = pandas.DataFrame([features_values_dict])
            data = data.append(input_df, sort=False).apply(pandas.to_numeric, errors='ignore', axis=0)
            # Drop missing columns, expected only to be label column
            data = data.dropna(axis=1)
        else:
            data = pandas.DataFrame([features_values_dict])
    except Exception as e:
        print("Could not parse properly the input data" + " - error: " + str(e))
        return None, None, None

    # Use feature encoder
    non_numeric = data.select_dtypes(include='object')
    if not non_numeric.empty and easier_model.get_feature_encoder():          
        data, _ = easier.datasets.one_hot_encode_data(data, one_hot_encoder=easier_model.get_feature_encoder())
                
    # Use scaler
    input_data = data.values
    if easier_model.get_scaler():
        input_data = easier_model.get_scaler().transform(input_data).reshape(len(input_data),-1)
    
    # Predict
    predictions = easier_model.get_model().predict(input_data)

    # Decode predictions if classifier
    if easier_model.get_label_encoder():
        classes_pred = easier_model.get_label_encoder().inverse_transform(numpy.argmax(predictions, axis=1))
        confidences = [float(prediction[numpy.argmax(predictions, axis=1)][0]) for prediction in predictions]
        return confidences, classes_pred, predictions
    else:
        return None, None, predictions

if __name__== "__main__":
    parser = argparse.ArgumentParser(description='EASIER Model Docker Image')
    parser.add_argument('--port', type=int, help='Port to access model.', default=5000)
    parser.add_argument('--models_path', type=str, help='Path to read models files. Default is /dockerized_model', default='./dockerized_model')
    args = parser.parse_args() # default to system args

    # Initializations
    port = os.getenv("port", args.port)
    easier_user = os.getenv("easier_user")
    easier_password = os.getenv("easier_password")
    
    ignore_cert_warning = os.getenv("ignore_cert_warning")
    ignore_cert_warning = True if str(ignore_cert_warning).lower() == 'true' else False

    minio_region = os.getenv("minio_region")
    
    minio_secure = os.getenv("minio_secure")
    minio_secure = True if str(minio_secure).lower() == 'true' else False

    easier_url = os.getenv("easier_url")
    easier = EasierSDK(easier_user=easier_user, easier_password=easier_password, easier_url=easier_url, minio_secure=minio_secure, minio_region=minio_region, ignore_cert_warning=ignore_cert_warning)

    repo = os.getenv("repo")
    category = os.getenv("category")
    model_name = os.getenv("model_name")
    experimentID = os.getenv("experimentID")

    easier_model = easier.models.get_model(repo_name=repo, category=category, model_name=model_name, experimentID=experimentID)    

    app.run(host="0.0.0.0", port=port, debug=False, threaded=True)




#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import sys
sys.path.append('..')
sys.path.append('.')

import time
import json

from easierSDK.classes.categories import Categories

class DatasetMetadata():
    """Class to store the dataset's metadata information.
    """

    category = ''
    name = ''
    last_modified = ''
    size = 0
    description = ''
    version = 0
    features = {}
    row_number = 0
    dataset_type = ''
    file_extension = ''

    def __init__(self, f:dict=None):
        """Constructor for the Dataset Metadata Class.

        Args:
            f (dict, optional): Dictionary with the dataset's metadata information.
        """
        if f is not None:
            if 'category' in f: 
                try:
                    self.category = Categories(f['category'])
                except:
                    print("[DatasetMetadata] Category " + str(f['category']) + " not recognized. Using 'misc' as categoriy")
                    self.category = Categories.MISC
            if 'name' in f: self.name = f['name']
            if 'last_modified' in f: 
                try:
                    self.last_modified = time.strftime('%H:%M:%S - %d/%m/%Y', f['last_modified'])
                except:
                    # print("[DatasetMetadata] Couldn't parse time in 'last-modified'. Keeping as str")
                    self.last_modified = f['last_modified']
            if 'size' in f: self.size = f['size']
            if 'description' in f: self.description = f['description']
            if 'version' in f: self.version = f['version']
            if 'features' in f: self.features = f['features']
            if 'row_number' in f: self.row_number = f['row_number']
            if 'dataset_type' in f: self.dataset_type = f['dataset_type']
            if 'file_extension' in f: self.file_extension = f['file_extension']
        else:
            pass

    # def seria
    def pretty_print(self):
        """Print the metadata information.
        """
        if isinstance(self.category, str):
            category = Categories[self.category.upper()]
        else:
            category = self.category.value
        row_format ="{:<30}" * 2
        print(row_format.format(*['Category:', category]))
        print(row_format.format(*['Name:', self.name]))
        print(row_format.format(*['Size:', self.size]))
        print(row_format.format(*['Description:', self.description]))
        print(row_format.format(*['Last modified:', str(self.last_modified)]))
        print(row_format.format(*['Version:', self.version]))
        print(row_format.format(*['Row number:', self.row_number]))
        print(row_format.format(*['Features:', str(self.features)]))
        print(row_format.format(*['Dataset type:', self.dataset_type]))
        print(row_format.format(*['File extension:', self.file_extension]))

    def dump_to_file(self, path:str):
        """Dump metadata information to a file.

        Args:
            path (str): Path to store the file with the metadata information.
        """
        if isinstance(self.category, str):
            self.category = Categories[self.category.upper()]
        else:
            self.category = self.category.value
        with open(path+'/metadata.json', 'w') as f:
            f.write(json.dumps(self.__dict__, default=lambda o: str(o.value), 
            sort_keys=True, indent=4))




import os
import kubernetes
import yaml
from kubernetes import client, config
from kubernetes.client.rest import ApiException
import string
import random
import joblib
import tempfile
import json

from easierSDK.classes.categories import Categories
from easierSDK.classes import constants
from easierSDK.classes.easier_model import EasierModel
from easierSDK.classes.training_metadata import TrainingMetadata
from easierSDK.training import easier_training_config_map
from easierSDK.training import easier_training_deployment
from easierSDK.training import easier_training_service 
from easierSDK.training import easier_training_ingress 

class TrainingAPI():
    """Class to control the Training API of EasierSDK.
    """

    _easier_user = None
    _easier_password = None

    def __init__(self, easier_user, easier_password, minio_client, my_public_repo, my_private_repo, easier_instance=None):
        """Constructor for the TrainingAPI.

        Args:
            minio_client (Minio): Minio client object with user session initialized.
            my_public_repo (str): Name of the public bucket of the user.
            my_private_repo (str): Name of the private bucket of the user.
        """
        self._easier_user = easier_user
        self._easier_password = easier_password
        self.minio_client = minio_client
        self.my_public_repo = my_public_repo
        self.my_private_repo = my_private_repo
        self._easier_instance = easier_instance
        
    def _id_generator(self, size=16, chars=string.ascii_lowercase + string.digits):
        """Generates random chars and returns a string.

        Args:
            size (int, optional): Size of generated string. Defaults to 16.
            chars (list, optional): types of chars to use for the generation. Defaults to string.ascii_lowercase+string.digits.

        Returns:
            str: string with randomly generated chars
        """
        return ''.join(random.choice(chars) for _ in range(size))

    def get_training_data(self, repo_path:str=None):
        """Loads training data from repository.

        Args:
            repo_path (str, optional): Path in repository to find training data. Defaults to None to read environment variable "model_path".

        Returns:
            tuple: tuple of size 4 with: x_train, y_train, x_test, y_test
        """
        
        # Pickle x and y from minio
        x = None
        y = None
        x_test = None
        y_test = None
        with tempfile.TemporaryDirectory() as path:
            if repo_path is None: repo_path = os.getenv("data_path", '/train/data')

            object_list = self.minio_client.list_objects(self.my_private_repo, prefix=repo_path, recursive=True)
            has_items = False
            # 3. Download
            for obj in object_list:
                if not obj.is_dir:
                    has_items = True
                    self.minio_client.fget_object(self.my_private_repo, obj.object_name, path+'/'+obj.object_name)
            if not has_items:
                print('[ERROR] Could not find file. Please check parameters and try again.')
                return False
            try:
                x = joblib.load(path+'/'+repo_path+'/x.joblib')
            except Exception as e:
                print("[ERROR] Loading training data (x): " + str(e))
            try:
                y = joblib.load(path+'/'+repo_path+'/y.joblib')
            except Exception as e:
                print("[ERROR] Loading training data (y): " + str(e))
            try:
                x_test = joblib.load(path+'/'+repo_path+'/x_test.joblib')
            except Exception as e:
                print("[ERROR] Loading validation data (x_test): " + str(e))
            try:
                y_test = joblib.load(path+'/'+repo_path+'/y_test.joblib')
            except Exception as e:
                print("[ERROR] Loading validation data (y_test): " + str(e))

        return x, y, x_test, y_test

    def clear_training_data(self, repo_path:str, repo_name:str):
        """Removes training data from repository.

        Args:
            repo_path (str): path in repository of the data
            repo_name (str): name of the repository
        """
        # Remove a prefix recursively.
        for obj in self.minio_client.list_objects(repo_name, repo_path, recursive=True):
            try:
                self.minio_client.remove_object(repo_name, obj.object_name)
            except Exception as e:
                print("[ERROR] When deleting object: ", str(e))

    def create_training(self, x, y, easier_model:EasierModel, training_metadata:TrainingMetadata=None, x_test=None, y_test=None, public=False, resource_name=None):
        """Creates a training deployment on EASIER platform with the model given in easier_model according to its training parameters.

        Args:
            x (list): training data in a list-like object
            y (list): labels or prediction targets in a list-like object
            easier_model (EasierModel): Model to be trained encapsulated in an EasierModel variable.
            training_metadata (TrainingMetadata, optional): Metadata of the training process. Defaults to None to take the training_metadata parameter of the easier_model.
            x_test (list): validation data in a list-like object
            y_test (list): validation labels or targets in a list-like object
            public (bool, optional): whether this model is going to be public for other users or not. Defaults to False.
            namespace ([type], optional): kubernetes namespace where the deployment is going to be launched. Defaults to None.

        Returns:
            [str]: name of the kubernetes deployment identifier
        """
        # Test if model can be loaded
        if easier_model.get_model() is None:
            print("ERROR: There is no model to train")
            return None
        
        if resource_name is None: 
            random_name = self.id_generator(size=5)
        else:
            random_name = resource_name

        minio_path = 'train/' + random_name

        # Pickle x and y and upload to minio
        with tempfile.TemporaryDirectory() as path:
            filename_x = os.path.join(path, 'x.joblib')
            filename_y = os.path.join(path, 'y.joblib')
            joblib.dump(x, filename_x, compress=True)
            joblib.dump(y, filename_y, compress=True)

            filename_x_test = os.path.join(path, 'x_test.joblib')
            filename_y_test = os.path.join(path, 'y_test.joblib')
            joblib.dump(x_test, filename_x_test, compress=True)
            joblib.dump(y_test, filename_y_test, compress=True)

            if training_metadata: easier_model.set_training_metadata(training_metadata)
            
            # Upload model and set it to training state
            self._easier_instance.models.upload(easier_model, public=public, training=True)
            
            if public:
                bucket = self.my_public_repo
            else:    
                bucket = self.my_private_repo
            
            # Create bucket if doesn't exist
            if not self.minio_client.bucket_exists(bucket): self.minio_client.make_bucket(bucket, location='es')

            # Upload training data
            for f in os.listdir(path):
                try:    
                    file_path = (minio_path + "/" + f)
                    a, b =self.minio_client.fput_object(bucket, file_path, path + '/' + f)
                except Exception as ex:
                    print('[ERROR] Unknown error uploading file {}: {}'.format(f, ex))
                    return None

        namespace = self.namespace  
        
        # Enter a context with an instance of the API kubernetes.client
        with kubernetes.client.ApiClient() as api_client:
            # Create an instance of the API class
            api_instance = kubernetes.client.CoreV1Api(api_client)
            
            # api_instance.create_namespaced_persistent_volume_claim()
            
            # config_map = None
            # with open(os.path.join(os.path.dirname(__file__), "easier-training-config_map.yaml")) as f:
            config_map = yaml.safe_load(easier_training_config_map.config_map)

            config_map['metadata']['name'] += '-' + random_name
            config_map['data']['easier_user'] = self._easier_user
            config_map['data']['easier_password'] = self._easier_password
            config_map['data']['repo'] = bucket
            config_map['data']['category'] = easier_model.metadata.category.value
            config_map['data']['model_name'] = easier_model.metadata.name
            config_map['data']['experimentID'] = str(easier_model.metadata.version)
            config_map['data']['data_path'] = minio_path
            config_map['data']['ignore_cert_warning'] = str(self._easier_instance.ignore_cert_warning)
            config_map['data']['minio_region'] = self._easier_instance.minio_region
            config_map['data']['minio_secure'] = str(self._easier_instance.minio_secure)
            config_map['data']['easier_url'] = self._easier_instance.easier_url
            
            try:
                api_response = api_instance.create_namespaced_config_map(namespace, config_map, pretty='true')
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->create_namespaced_config_map: %s\n" % e)

            api_instance = kubernetes.client.BatchV1Api(api_client)
            # deployment = None
            # with open(os.path.join(os.path.dirname(__file__), "easier-training-deployment.yaml")) as f:
            deployment = yaml.safe_load(easier_training_deployment.deployment)
        
            deployment['metadata']['name'] += '-' + random_name
            deployment['metadata']['labels']['app'] += '-' + random_name
            deployment['spec']['template']['metadata']['labels']['app'] += '-' + random_name
            deployment['spec']['template']['spec']['containers'][0]['name'] += '-' + random_name
            deployment['spec']['template']['spec']['containers'][0]['envFrom'][0]['configMapRef']['name'] += '-' + random_name
    
            try:
                api_response = api_instance.create_namespaced_job(namespace, deployment, pretty='true')
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->create_namespaced_pod: %s\n" % e)

        print("Resources identifier: " + random_name)
        return random_name

    def clear_training_deployment(self, deployment_identifier:str):
        """Deletes the deployment resources of a training.

        Args:
            deployment_identifier (str): identifier given when a training was created
            namespace (str): namespace in which to delete the deployment
        """
        namespace = self.namespace
        # Enter a context with an instance of the API kubernetes.client
        with kubernetes.client.ApiClient() as api_client:
            # Create an instance of the API class
            api_instance = kubernetes.client.CoreV1Api(api_client)

            try:
                config_map = yaml.safe_load(easier_training_config_map.config_map)
            
                config_map['metadata']['name'] += '-' + deployment_identifier
                api_response = api_instance.delete_namespaced_config_map(config_map, namespace)
                # print(api_response)
            except ApiException as e:
                print("Exception when calling CoreV1Api->delete_namespaced_config_map: %s\n" % e)

            # try:
            #     deployment = yaml.safe_load(easier_training_deployment.deployment)
            
            #     deployment['metadata']['name'] += '-' + deployment_identifier
            #     api_response = api_instance.delete_namespaced_pod(deployment, namespace)
            #     # print(api_response)
            # except ApiException as e:
            #     print("Exception when calling CoreV1Api->delete_namespaced_pod: %s\n" % e)

            api_instance = kubernetes.client.BatchV1Api(api_client)
            deployment = yaml.safe_load(easier_training_deployment.deployment)        
            deployment['metadata']['name'] += '-' + deployment_identifier
            try:
                api_response = api_instance.delete_namespaced_job(deployment, self.namespace)
                # print(api_response)
            except ApiException as e:
                print("Exception when calling BatchV1Api->delete_namespaced_job: %s\n" % e)

    def _get_resource_names(self, resource_list):
        resource_names = []
        for resource in resource_list:
            resource_names.append(resource.metadata.name)
        return resource_names

    def print_training_logs(self, deployment_identifier):

        # Enter a context with an instance of the API kubernetes.client
        with kubernetes.client.ApiClient() as api_client:
            deployment = yaml.safe_load(easier_training_deployment.deployment)        
            deployment['metadata']['name'] += '-' + deployment_identifier

            api_instance = kubernetes.client.CoreV1Api(api_client)
            name = deployment['metadata']['name'] # name of the Pod
            pretty = 'true' 
            
            try:
                api_response = api_instance.list_namespaced_pod(self.namespace, pretty='true')
                pods_list = self._get_resource_names(api_response.items)
            except ApiException as e:
                print("Exception when calling CoreV1Api->list_namespaced_pod: %s\n" % e)
                
            if len(pods_list) == 0:
                print("No trainings found with identifier \"" + deployment_identifier + "\".")
            else:
                try: 
                    pod_name = [string for string in pods_list if name in string]
                    api_response = api_instance.read_namespaced_pod_log(pod_name[0], self.namespace, pretty=pretty)
                    print(api_response)
                except ApiException as e:
                    print("Exception when calling CoreV1Api->read_namespaced_pod_log: %s\n" % e)